"""coBib's UI components.

This module provides various UI-related components such as widgets and other utilities.

.. warning::

   This module makes no API stability guarantees! With the `cobib.ui.tui.TUI` being based on
   [`textual`](https://textual.textualize.io/) which is still in very early stages of its
   development, breaking API changes in this module might be released as part of coBib's feature
   releases. You have been warned.
"""

from .entry_view import EntryView as EntryView
from .input_screen import InputScreen as InputScreen
from .list_view import ListView as ListView
from .log_screen import LogScreen as LogScreen
from .logging_handler import LoggingHandler as LoggingHandler
from .main_content import MainContent as MainContent
from .motion_key import MotionKey as MotionKey
from .note_view import NoteView as NoteView
from .preset_filter_screen import PresetFilterScreen as PresetFilterScreen
from .search_view import SearchView as SearchView
from .selection_filter import SelectionFilter as SelectionFilter



"""coBib's search results viewer widget.

This widget gets produces by `cobib.commands.search.SearchCommand.render_textual`.
It subclasses textual's built-in `Tree` widget.

.. warning::

   This module makes no API stability guarantees! Refer to `cobib.ui.components` for more details.
"""

from __future__ import annotations

from typing import ClassVar, Union

from rich.text import Text
from textual.binding import Binding
from textual.widgets import Tree
from typing_extensions import override

from .motion_key import MotionKey


class SearchView(Tree[Union[str, Text]]):
    """coBib's search results viewer widget."""

    id = "cobib-search-view"

    BINDINGS: ClassVar[list[Binding | tuple[str, str] | tuple[str, str, str]]] = [
        Binding(
            "space",
            "toggle_node",
            "Toggle",
            tooltip="Toggles one fole level of the search result",
        ),
        Binding(
            "backspace",
            "toggle_all",
            "Toggle All",
            tooltip="Toggles all fold levels of the search result",
        ),
        Binding(
            "j",
            "motion('down', 'cursor_down')",
            "Down",
            tooltip="Moves one row down",
            show=False,
        ),
        Binding(
            "k",
            "motion('up', 'cursor_up')",
            "Up",
            tooltip="Moves one row up",
            show=False,
        ),
        Binding(
            "h",
            "motion('left', 'cursor_left')",
            "Left",
            tooltip="Moves to the left",
            show=False,
        ),
        Binding(
            "l",
            "motion('right', 'cursor_right')",
            "Right",
            tooltip="Moves to the right",
            show=False,
        ),
        Binding(
            "down",
            "motion('down', 'cursor_down')",
            "Down",
            tooltip="Moves one row down",
            show=False,
        ),
        Binding(
            "up",
            "motion('up', 'cursor_up')",
            "Up",
            tooltip="Moves one row up",
            show=False,
        ),
        Binding(
            "left",
            "motion('left', 'cursor_left')",
            "Left",
            tooltip="Moves to the left",
            show=False,
        ),
        Binding(
            "right",
            "motion('right', 'cursor_right')",
            "Right",
            tooltip="Moves to the right",
            show=False,
        ),
        Binding(
            "home",
            "motion('home', 'scroll_top')",
            "Home",
            tooltip="Jumps to the first row",
            show=False,
        ),
        Binding(
            "end",
            "motion('end', 'scroll_bottom')",
            "End",
            tooltip="Jumps to the last row",
            show=False,
        ),
        Binding(
            "pageup",
            "motion('pageup', 'page_up')",
            "Page Up",
            tooltip="Moves up by one page",
            show=False,
        ),
        Binding(
            "pagedown",
            "motion('pagedown', 'page_down')",
            "Page Down",
            tooltip="Moves down by one page",
            show=False,
        ),
    ]
    """
    | Key(s) | Description |
    | :- | :- |
    | space | Toggle the expand/collapsed space of the current item. |
    | backspace | Toggle the expand/collapsed space of the current item and all its children. |
    | j, down | Moves one row down. |
    | k, up | Moves one row up. |
    | h, left | Moves to the left. |
    | l, right | Moves to the right. |
    | PageDown | Moves one page down. |
    | PageUp | Moves one page up. |
    | End | Moves to the bottom of the tree. |
    | Home | Moves to the top of the tree. |
    """

    @override
    def on_mount(self) -> None:
        self.show_root = False

    def action_motion(self, key: str, action: str) -> None:
        """Action to move the cursor.

        Under the hood, this delegates to the respective built-in cursor motion methods.
        However, this method also posts a `cobib.ui.components.motion_key.MotionKey` event.

        Args:
            key: the key which was pressed.
            action: the built-in action to trigger.
        """
        func = getattr(self, f"action_{action}", None)
        if func is not None:
            func()
        self.post_message(MotionKey(key))

    def action_toggle_all(self) -> None:
        """Toggles the expansion of the current node and all of its children recursively."""
        if self.cursor_node is not None:
            self.cursor_node.toggle_all()

    def get_current_label(self) -> str | None:
        """Gets the label of the entry currently under the cursor.

        Returns:
            The label of the entry currently under the cursor.
        """
        if self.cursor_node is None:
            return None
        return str(self.cursor_node.data)

    def jump_to_label(self, label: str) -> None:
        """Jumps to the requested label.

        Args:
            label: the label to jump to.

        Raises:
            KeyError: when the label was not found in the current view.
        """
        for child in self.root.children:
            if child.data == label:
                self.select_node(child)
                self.scroll_to_node(child, animate=True)
                break
        else:
            raise KeyError



"""coBib's note viewer widget.

This widget gets used to display and edit the contents of the note associated with the currently
selected entry.

.. warning::

   This module makes no API stability guarantees! Refer to `cobib.ui.components` for more details.
"""

from __future__ import annotations

import logging
from typing import ClassVar

from textual.binding import Binding
from textual.reactive import Reactive
from textual.widgets import TextArea
from textual.widgets.text_area import Edit, EditResult
from typing_extensions import override

from cobib.config import config
from cobib.database import Database

from .main_content import MainContent

LOGGER = logging.getLogger(__name__)
"""@private module logger."""


class NoteView(TextArea):
    """coBib's note viewer widget."""

    DEFAULT_CSS = """
        NoteView {
            height: 1fr;
            width: 1fr;
            grid-size: 1 2;
            border: tall $accent;
        }
    """

    BINDINGS: ClassVar[list[Binding | tuple[str, str] | tuple[str, str, str]]] = [
        Binding("escape", "escape", "Escape"),
        Binding("ctrl+r", "reset", "Reset"),
        Binding("ctrl+s", "save", "Save"),
        Binding("ctrl+x", "external", "External edit"),
    ]
    """
    | Key(s) | Description |
    | :- | :- |
    | Escape | Unfocuses the text area. |
    | Ctrl+r | Reset any unsaved changes. |
    | Ctrl+s | Saves the text area contents. |
    | Ctrl+x | Opens the text area content in an external editor. |
    """

    PLACEHOLDER = (
        "Here you can view and edit the `note` associated with the current entry.\n"
        "Quickstart:\n"
        "  - hit `Enter` to load the associated note\n"
        "  - hit `n` to focus the text area and start editing\n"
        "  - hit `Escape` to unfocus the text area\n"
        "  - hit `Ctrl+s` to save the note\n"
        "  - hit `Ctrl+r` to discard any unsaved changes\n"
        "  - hit `Ctrl+x` to open the note for editing in an external editor\n"
    )

    unsaved: Reactive[bool] = Reactive(False, compute=False)
    """A reactive boolean indicating whether the view contains unsaved changes."""

    @override
    def on_mount(self) -> None:
        self.text = NoteView.PLACEHOLDER
        self.border_title = "Note"
        self.read_only = True

    @override
    def load_text(self, text: str) -> None:
        super().load_text(text)
        self.unsaved = False

    @override
    def edit(self, edit: Edit) -> EditResult:
        result = super().edit(edit)
        self.unsaved = True
        return result

    def watch_has_focus(self, value: bool) -> None:
        """Watches whether this widget has focus.

        Args:
            value: the new focus value.
        """
        super().watch_has_focus(value)
        self._update_border(value)
        if not value:
            self.read_only = True

    def watch_unsaved(self, value: bool) -> None:
        """Watches the state of `unsaved`.

        Args:
            value: the new value assigned to `unsaved`.
        """
        self.border_subtitle = "UNSAVED" if value else ""
        self._update_border(True)

    def _update_border(self, has_focus: bool) -> None:
        """Updates the border style depending on the state of the widget.

        Args:
            has_focus: whether the widget currently has focus (see also `watch_has_focus`).
        """
        if has_focus:
            if self.unsaved:
                self.styles.border = ("tall", config.theme.css_variables["warning"])
            else:
                self.styles.border = ("tall", config.theme.css_variables["success"])
        else:  # noqa: PLR5501
            if self.unsaved:
                self.styles.border = ("tall", config.theme.css_variables["error"])
            else:
                self.styles.border = ("tall", config.theme.css_variables["accent"])

    def action_escape(self) -> None:
        """The `Esc` key action to unfocus this widget."""
        main = self.app.query_one(MainContent)
        main_content = None if main.current is None else main.get_child_by_id(main.current)
        self.screen.set_focus(main_content)

    def action_reset(self) -> None:
        """The `Ctrl+r` key action to reset any unsaved changes."""
        label = self.border_title
        if not isinstance(label, str):
            raise RuntimeError("Expecting to extract a label from the border title!")

        LOGGER.info(f"Discarding any unsaved changes to the note of entry '{label}'.")
        self.load_note(label, escape=True, force_reload=True)

    def action_save(self) -> None:
        """The `Ctrl+s` key action to save the current changes."""
        self._save_contents("_inline")

    def action_external(self) -> None:
        """The `Ctrl+x` key action to open the note in an external editor."""
        self._save_contents("edit")

    def load_note(
        self,
        label: str,
        *,
        edit: bool = False,
        escape: bool = False,
        force_reload: bool = False,
    ) -> None:
        """Loads a note into this widget.

        Args:
            label: the label of the `cobib.database.entry.Entry` whose note to load.
            edit: whether to immediately focus this widget and start editing the contents (`True`)
                or leave the widget in read-only mode (`False`, default).
            escape: whether to trigger `action_escape` after loading the note.
            force_reload: whether to discard any unsaved changes.
        """
        if self.border_title != label or force_reload:
            if self.unsaved:
                msg = (
                    "You have unsaved changes on your open note! You must save or reset them before"
                    " being able to load a different note!\n"
                    "To do so, focus the note area (by clicking or navigating to the corresponding "
                    "entry and hitting 'n') and hit 'Ctrl+s' to save or 'Ctrl+r' to reset the note."
                )
                self.notify(msg, title="Error", severity="error", timeout=5)
                return

            self.border_title = label
            self.border_subtitle = None

            from cobib.commands import NoteCommand

            note_file = NoteCommand.note_path(Database()[label])

            if note_file.path.exists():
                self.text = open(note_file.path, "r", encoding="utf-8").read()
            else:
                self.text = ""

        self.unsaved = False

        if escape:
            self.action_escape()

        elif edit:
            self.read_only = False
            self.focus()

    def _save_contents(self, action: str) -> None:
        """Saves the contents of the widget into the note file on disk.

        Args:
            action: the `action` argument to the `cobib.commands.note.NoteCommand`.
        """
        label = self.border_title
        if not isinstance(label, str):
            raise RuntimeError("Expecting to extract a label from the border title!")

        from cobib.commands import NoteCommand

        note_file = NoteCommand.note_path(Database()[label])

        with open(note_file.path, "w", encoding="utf-8") as file:
            file.write(self.text)

        NoteCommand(label, action).execute()

        self.load_note(label, escape=True, force_reload=True)



"""coBib's lint command.

This command allows you to lint your database.
"""

from __future__ import annotations

import argparse
import logging
from io import StringIO

from rich.console import ConsoleRenderable
from rich.text import Text
from typing_extensions import override

from cobib.database import Database
from cobib.utils.rel_path import RelPath

from .base_command import Command

LOGGER = logging.getLogger(__name__)
"""@private module logger."""


class LintFormatter(logging.Formatter):
    """A custom logging.Formatter."""

    def __init__(self, *args, **kwargs) -> None:  # type: ignore[no-untyped-def]
        """Initializes a LintFormatter."""
        super().__init__(*args, **kwargs)

        self.dirty_entries: set[str] = set()

        from cobib.config import config

        self._database_path = RelPath(config.database.file)

        with open(self._database_path.path, "r", encoding="utf-8") as database:
            self._raw_database = database.readlines()

    def format(self, record: logging.LogRecord) -> str:
        """Format's the LogRecord.

        This custom Formatter uses the LogRecord's attributes to determine from which line of the
        raw database a formatting information was raised. The corresponding line number is used in
        conjunction with the actual message of the LogRecord for the formatting.

        Args:
            record: the LogRecord to be formatted.

        Returns:
            A string encoding the LogRecord's information.
        """
        try:
            entry = record.entry  # type: ignore[attr-defined]
            field = record.field  # type: ignore[attr-defined]
            self.dirty_entries.add(entry)
            raw_db = enumerate(self._raw_database)
            _, line = next(raw_db)
            while not line.startswith(entry):
                _, line = next(raw_db)
            while not line.strip().startswith(field):
                line_no, line = next(raw_db)
            return f"{self._database_path}:{line_no+1} {record.getMessage()}"
        except AttributeError:
            return ""


class LintCommand(Command):
    """The lint Command.

    This command can parse the following arguments:

        * `-f`, `--format`: if specified, the database will be formatted to automatically resolve
            all lint messages.
    """

    name = "lint"

    @override
    def __init__(self, *args: str) -> None:
        super().__init__(*args)

        self._lint_messages: list[str] = []

    @override
    @classmethod
    def init_argparser(cls) -> None:
        parser = argparse.ArgumentParser(
            prog="lint", description="Lint subcommand parser.", exit_on_error=True
        )
        parser.add_argument(
            "-f",
            "--format",
            action="store_true",
            help="Automatically format database to conform with linter.",
        )
        cls.argparser = parser

    @override
    def execute(self) -> None:
        output = StringIO()

        handler = logging.StreamHandler(output)
        handler.setLevel(logging.INFO)
        handler.addFilter(logging.Filter("cobib.database.entry"))

        formatter = LintFormatter()
        handler.setFormatter(formatter)

        root_logger = logging.getLogger()
        if root_logger.getEffectiveLevel() > logging.INFO:
            # overwriting all existing handlers with this local one
            root_logger.handlers = [handler]
            root_logger.setLevel(logging.INFO)
        else:
            # appending new handler
            root_logger.addHandler(handler)

        # trigger database reading to cause lint messages upon entry-construction
        Database.read(bypass_cache=True)

        self._lint_messages = output.getvalue().split("\n")

        root_logger.removeHandler(handler)

        if all(not msg for msg in self._lint_messages):
            self._lint_messages = ["Congratulations! Your database triggers no lint messages."]

        elif self.largs.format:
            for label in formatter.dirty_entries:
                # we exploit the rename method to register all dirty entries for re-writing
                Database().rename(label, label)

            Database.save()
            self.git()

            self._lint_messages.insert(
                0, "The following lint messages have successfully been resolved:"
            )

    @override
    def render_porcelain(self) -> list[str]:
        return self._lint_messages

    @override
    def render_rich(self) -> ConsoleRenderable:
        text = Text("\n".join(self._lint_messages))
        return text



"""coBib's Edit command.

This command can be used to manually edit database entries in their easily-readable YAML format.
To get started, simply type:
```
cobib edit <label>
```
which will open the YAML-formatted version of the specified `cobib.database.Entry` for editing.

You can configure which editor will be used via the `cobib.config.config.EditCommandConfig.editor`
setting which will default to using your `$EDITOR` environment setting (and fall back to `vim` if
that is not set).

You can even add entirely new entries to the database by specifying an unused entry label *and*
adding the `--add` command-line argument:
```
cobib edit --add <new label>
```
This entry will be entirely empty except for the one field which is always present:
    * `ENTRYTYPE`: set to the default value configured via
      `cobib.config.config.EditCommandConfig.default_entry_type`.

If you change the label of the entry during editing, the value of the
`cobib.config.config.EditCommandConfig.preserve_files` setting (added in v4.1.0) determines whether
the associated files will be renamed automatically. This defaults to `False`, meaning that they
*will* be renamed. You can overwrite the value of this setting at runtime with the
`--preserve-files` and `--no-preserve-files` arguments, respectively.
I.e. the following will **not** rename your files:
```
cobib edit --preserve-files <label>
```
While this command will always rename them:
```
cobib edit --no-preserve-files <label>
```

### TUI

You can also trigger this command from the `cobib.ui.tui.TUI`.
By default, it is bound to the `e` key.
If you want to add a new entry manually, you will have to enter the prompt (defaults to `:`) and
then type out the command mentioned above:
```
:edit --add <new label>
```
"""

from __future__ import annotations

import argparse
import logging
import os
import tempfile
from pathlib import Path

from typing_extensions import override

from cobib.config import Event, config
from cobib.database import Database, Entry
from cobib.parsers.yaml import YAMLParser
from cobib.utils.context import get_active_app
from cobib.utils.rel_path import RelPath

from .base_command import Command

LOGGER = logging.getLogger(__name__)
"""@private module logger."""


class EditCommand(Command):
    """The Edit Command.

    This command can parse the following arguments:

        * `label`: the label of the entry to edit.
        * `-a`, `--add`: if specified, allows adding a new entry for a non-existent label. The
          default entry type of this new entry can be configured via
          `cobib.config.config.EditCommandConfig.default_entry_type`.
        * `--preserve-files`: skips the renaming of any associated files in case you manually rename
          the entry label during editing. This overwrites the
          `cobib.config.config.EditCommandConfig.preserve_files` setting.
        * `--no-preserve-files`: does NOT skip the renaming of any associated files in case you
          manually rename the entry label during editing. This overwrites the
          `cobib.config.config.EditCommandConfig.preserve_files` setting.
    """

    name = "edit"

    @override
    def __init__(self, *args: str) -> None:
        super().__init__(*args)

        self.new_entry: Entry
        """A `cobib.database.Entry` instance edited by this command."""

    @override
    @classmethod
    def init_argparser(cls) -> None:
        parser = argparse.ArgumentParser(
            prog="edit", description="Edit subcommand parser.", exit_on_error=True
        )
        parser.add_argument("label", type=str, help="label of the entry")
        parser.add_argument(
            "-a",
            "--add",
            action="store_true",
            help="if specified, will add a new entry for unknown labels",
        )
        preserve_files_group = parser.add_mutually_exclusive_group()
        preserve_files_group.add_argument(
            "--preserve-files",
            action="store_true",
            default=None,
            help="do NOT rename associated files",
        )
        preserve_files_group.add_argument(
            "--no-preserve-files",
            dest="preserve_files",
            action="store_false",
            default=None,
            help="rename associated files",
        )
        cls.argparser = parser

    @override
    def execute(self) -> None:
        LOGGER.debug("Starting Edit command.")

        Event.PreEditCommand.fire(self)

        yml = YAMLParser()

        bib = Database()

        try:
            entry = bib[self.largs.label]
            entry_text = yml.dump(entry)
            if self.largs.add:
                LOGGER.warning(
                    "Entry '%s' already exists! Ignoring the `--add` argument.", self.largs.label
                )
                self.largs.add = False
        except KeyError:
            # No entry for given label found
            if self.largs.add:
                # add a new entry for the unknown label
                entry = Entry(
                    self.largs.label,
                    {"ENTRYTYPE": config.commands.edit.default_entry_type},
                )
                entry_text = yml.dump(entry)
            else:
                msg = (
                    f"No entry with the label '{self.largs.label}' could be found."
                    "\nUse `--add` to add a new entry with this label."
                )
                LOGGER.error(msg)
                return
        if entry_text is None:
            # No entry found to be edited. This should never occur unless the YAMLParser experiences
            # an unexpected error.
            LOGGER.error(
                f"Encountered an unexpected case while trying to edit {self.largs.label} which "
                "might result from a problem with the YAML parser."
            )
            return

        new_text = self.edit(entry_text)

        if entry_text == new_text and not self.largs.add:
            LOGGER.info("No changes detected.")
            return

        parsed = yml.parse(new_text)
        self.new_entry = next(iter(parsed.values()))

        bib.update({self.new_entry.label: self.new_entry})

        preserve_files = config.commands.edit.preserve_files
        if self.largs.preserve_files is not None:
            preserve_files = self.largs.preserve_files
        LOGGER.info("Associated files will%s be preserved.", "" if preserve_files else " not")

        if self.new_entry.label != self.largs.label:
            bib.rename(self.largs.label, self.new_entry.label)
            if not preserve_files:
                new_files = []
                for file in self.new_entry.file:
                    path = RelPath(file)
                    if path.path.stem == self.largs.label:
                        LOGGER.info("Also renaming associated file '%s'.", str(path))
                        target = RelPath(path.path.parent / f"{self.new_entry.label}.pdf")
                        if target.path.exists():
                            LOGGER.warning("Found conflicting file, not renaming '%s'.", str(path))
                        else:
                            path.path.rename(target.path)
                            new_files.append(str(target))
                            continue
                    new_files.append(file)
                self.new_entry.file = new_files

        Event.PostEditCommand.fire(self)
        bib.save()

        self.git()

        msg = f"'{self.largs.label}' was successfully edited."
        LOGGER.info(msg)

    @staticmethod
    def edit(entry_text: str) -> str:
        """Spawns an editor in the terminal to edit the provided entry text.

        Args:
            entry_text: the entry in text form to edit.

        Returns:
            The new entry.
        """
        app = get_active_app()
        if app is None:
            return EditCommand._edit(entry_text)
        with app.suspend():
            return EditCommand._edit(entry_text)

    @staticmethod
    def _edit(entry_text: str) -> str:
        LOGGER.debug("Creating temporary file.")
        with tempfile.NamedTemporaryFile(mode="w+", prefix="cobib-", suffix=".yaml") as tmp_file:
            tmp_file_name = tmp_file.name
            tmp_file.write(entry_text)
            tmp_file.flush()
            LOGGER.debug('Starting editor "%s".', config.commands.edit.editor)
            status = os.system(config.commands.edit.editor + " " + tmp_file.name)
            assert status == 0
            LOGGER.debug("Editor finished successfully.")
            new_text = open(tmp_file.name, "r", encoding="utf-8").read()
        assert not Path(tmp_file_name).exists()
        return new_text



"""coBib's log screen.

This screen renders the log messages.

.. warning::

   This module makes no API stability guarantees! Refer to `cobib.ui.components` for more details.
"""

from __future__ import annotations

from typing import ClassVar

from textual.app import ComposeResult
from textual.binding import Binding
from textual.screen import ModalScreen
from textual.widgets import RichLog, Static
from typing_extensions import override


class LogScreen(ModalScreen[None]):
    """coBib's log screen."""

    BINDINGS: ClassVar[list[Binding | tuple[str, str] | tuple[str, str, str]]] = [
        ("z", "toggle_log", "log"),
    ]
    """
    | Key(s) | Description |
    | :- | :- |
    | z | Toggles the log screen. |
    """

    DEFAULT_CSS = """
        LogScreen {
            align: center bottom;
            offset-y: -1;
        }

        #caption {
            background: $surface;
        }

        #log {
            height: auto;
            max-height: 50%;
            background: $surface;
        }
    """

    rich_log = RichLog(id="log")
    """The internal `RichLog` widget in which the actual log messages are rendered."""

    @override
    def compose(self) -> ComposeResult:
        yield self.rich_log
        yield Static(
            "These are your log messages. Press 'z' to close this screen.",
            id="caption",
        )

    def action_toggle_log(self) -> None:
        """Toggles the log.

        Since this is the action of the `LogScreen`, it simply pops the screen.
        """
        self.app.pop_screen()



"""coBib's `PresetFilter` selection screen.

This screen renders a Textual-`OptionList` with all the preset filters configured via
`config.tui.preset_filters` for the user to choose from.

.. warning::

   This module makes no API stability guarantees! Refer to `cobib.ui.components` for more details.
"""

from __future__ import annotations

from typing import ClassVar

from textual.app import ComposeResult
from textual.binding import Binding
from textual.screen import ModalScreen
from textual.widgets import OptionList, Static
from typing_extensions import override

from cobib.config import config


class PresetFilter(OptionList):
    """coBib's custom `OptionList` which can be navigated with Vi-like bindings."""

    BINDINGS: ClassVar[list[Binding | tuple[str, str] | tuple[str, str, str]]] = [
        Binding("j", "cursor_down", "Down", show=False),
        Binding("k", "cursor_up", "Up", show=False),
    ]
    """
    | Key(s) | Description |
    | :- | :- |
    | j | Moves the cursor down. |
    | k | Moves the cursor up. |
    """


class PresetFilterScreen(ModalScreen[str]):
    """coBib's `PresetFilter` selection screen."""

    BINDINGS: ClassVar[list[Binding | tuple[str, str] | tuple[str, str, str]]] = [
        ("q", "abort", "Close"),
    ]
    """
    | Key(s) | Description |
    | :- | :- |
    | q | Aborts the selection. |
    """

    DEFAULT_CSS = """
        PresetFilterScreen {
            align: center middle;
        }

        #presets {
            width: 80%;
            max-height: 80%;
            background: $surface;
        }

        #help {
            color: $text-disabled;
            width: 80%;
            padding: 1 2;
            background: $surface;
        }
    """

    @override
    def compose(self) -> ComposeResult:
        yield Static(
            (
                "Select one of the preset filters configured via "
                "[yellow]config.tui.preset_filters[/yellow]. "
                "You can restore the default view by selecting the [yellow]NONE[/yellow] filter. "
                "You can abort the selection by pressing [yellow]q[/yellow]."
            ),
            id="help",
        )
        yield PresetFilter("NONE", *config.tui.preset_filters, id="presets")

    def action_abort(self) -> None:
        """Aborts the `PresetFilter` selection."""
        self.dismiss("")

    def on_option_list_option_selected(self, event: OptionList.OptionSelected) -> None:
        """Handles the `OptionList.OptionSelected` event.

        Args:
            event: the triggered event.
        """
        if event.option.prompt == "NONE":
            self.dismiss("list -r")
        else:
            self.dismiss(f"list {event.option.prompt}")



# mypy: warn-unused-ignores=False
"""coBib's YAML parser.

This parser leverages the [`ruamel.yaml`](https://pypi.org/project/ruamel.yaml/) library to convert
between `cobib.database.Entry` instances and YAML representations of their `dict`-like data
structure.

The parser is registered under the `-y` and `--yaml` command-line arguments of the
`cobib.commands.add.AddCommand`.

The following documentation is mostly inherited from the abstract interface
`cobib.parsers.base_parser`.
"""

from __future__ import annotations

import io
import logging
import sys
from collections import OrderedDict
from pathlib import Path
from typing import IO

from rich.console import Console
from rich.progress import track
from ruamel import yaml
from typing_extensions import override

from cobib.config import Event, config
from cobib.database.author import Author
from cobib.database.entry import Entry

from .base_parser import Parser

LOGGER = logging.getLogger(__name__)
"""@private module logger."""


class YAMLParser(Parser):
    """The YAML Parser."""

    name = "yaml"

    _yaml: yaml.YAML | None = None

    def __init__(self) -> None:
        """Initializes a YAMLParser."""
        if YAMLParser._yaml is None:
            # we need to lazily construct this in order to be able to respect the config setting
            YAMLParser._yaml = yaml.YAML(typ="safe", pure=not config.parsers.yaml.use_c_lib_yaml)
            YAMLParser._yaml.explicit_start = True  # type: ignore[assignment]
            YAMLParser._yaml.explicit_end = True  # type: ignore[assignment]
            YAMLParser._yaml.default_flow_style = False
            YAMLParser._yaml.register_class(Author)

    @override
    def parse(self, string: str | Path) -> dict[str, Entry]:
        string = Event.PreYAMLParse.fire(string) or string

        try:
            LOGGER.debug("Attempting to load YAML data from file: %s.", string)
            stream = open(string, "r", encoding="utf-8")
            bib = self._load_all(stream)
            stream.close()
        except (OSError, FileNotFoundError) as exc:
            try:
                LOGGER.debug("Attempting to load YAML data from string: %s.", string)
                stream = io.StringIO(string)  # type: ignore[arg-type]
                bib = self._load_all(stream)
                stream.close()
            except (TypeError, AttributeError):
                raise exc

        Event.PostYAMLParse.fire(bib)

        return bib

    def _load_all(self, stream: IO) -> dict[str, Entry]:  # type: ignore[type-arg]
        bib: dict[str, Entry] = OrderedDict()

        for entry in track(
            self._yaml.load_all(stream),  # type: ignore[union-attr]
            description="Reading database...",
            transient=True,
            console=Console(file=sys.stderr),
        ):
            for label, data in entry.items():
                actual_entry = Entry(label, data)
                if actual_entry.label in bib.keys():
                    LOGGER.warning(
                        "An entry with label '%s' was already encountered earlier on in the YAML "
                        "file! Please check the file manually as this cannot be resolved "
                        "automatically by coBib.",
                        actual_entry.label,
                    )
                bib[actual_entry.label] = actual_entry

        return bib

    @override
    def dump(self, entry: Entry) -> str | None:
        Event.PreYAMLDump.fire(entry)

        LOGGER.debug("Converting entry %s to YAML format.", entry.label)
        stream = io.StringIO()
        self._yaml.dump(  # type: ignore[union-attr]
            {entry.label: dict(sorted(entry.data.items()))}, stream=stream
        )
        string = stream.getvalue()

        string = Event.PostYAMLDump.fire(string) or string

        return string



"""coBib's Note command.

This command allows you to manipulate a note for an entry in your database. More precisely, it
allows you to show, edit, and delete the file linked to by the `cobib.database.entry.Entry.note`
field.

.. note::
   The *benefit* of the `cobib.database.entry.Entry.note` field over a file attached via the
   `cobib.database.entry.Entry.file` field is, that the note's contents are known to be plain-text
   and, thus, their content is included during the `cobib.commands.search.SearchCommand`.

The different actions that you can execute are very simple:

1. Show the note's contents:
   ```
   cobib note Label1 show
   ```

2. Edit the note's contents using `cobib.config.config.EditCommandConfig.editor`:
   ```
   cobib note Label1 edit
   ```

3. Delete the associated note:
   ```
   cobib note Label1 delete
   ```

When an entry does not have an associated note yet, the `edit` action will create a new empty note
for you. Its file location is determined as described in `NoteCommand.note_path`.

### TUI

You can also trigger this command from the `cobib.ui.tui.TUI`.
By default, it is bound to the `n` key which will load the note's content into the
`cobib.ui.components.note_view.NoteView` widget and start the editing process.
You can also preview the note's content using the `Enter` key.
The widget provides a few more features which it documents itself, so be sure to read that, too.
"""

from __future__ import annotations

import argparse
import logging
import os

from rich.console import ConsoleRenderable
from rich.syntax import Syntax
from typing_extensions import override

from cobib.config import Event, config
from cobib.database import Database, Entry
from cobib.utils.context import get_active_app
from cobib.utils.rel_path import RelPath

from .base_command import Command

LOGGER = logging.getLogger(__name__)
"""@private module logger."""


class NoteCommand(Command):
    """The Note Command.

    This command can parse the following arguments:

        * `label`: the label of the entry whose note to act upon.
        * `action`: the action to perform on the note. The following options are available:
            - `edit`: opens the note in an external editor.
            - `show`: prints the note contents.
            - `delete`: deletes the associated note.
    """

    name = "note"

    @override
    def __init__(self, *args: str) -> None:
        super().__init__(*args)

        self.note_content: str | None = None
        """The text content of the note. This may be `None` for a non-existent note."""

    @override
    @classmethod
    def init_argparser(cls) -> None:
        parser = argparse.ArgumentParser(
            prog="note", description="Note subcommand parser.", exit_on_error=True
        )
        parser.add_argument("label", type=str, help="label of the entry")
        parser.add_argument(
            "action",
            help="the action to perform on the note",
            choices=("edit", "show", "delete", "_inline"),
            default="edit",
            nargs="?",
        )
        cls.argparser = parser

    @staticmethod
    def note_path(entry: Entry) -> RelPath:
        """Generates the default path of a note file.

        Given an `cobib.database.entry.Entry`, the default path is determined as follows:
          - the filename is simply the `cobib.database.entry.Entry.label`
          - the filetype is configured via `cobib.config.config.NoteCommandConfig.default_filetype`
          - the location is adjacent to `cobib.config.config.DatabaseConfig.file`

        Alternatively, if the `cobib.database.entry.Entry.note` attribute is already set, that path
        is used unmodified.

        Args:
            entry: the `cobib.database.entry.Entry` whose note path to generate.

        Returns:
            The `cobib.utils.rel_path.RelPath` to the note file.
        """
        path_str = entry.note
        if path_str is None:
            filename = f"{entry.label}.{config.commands.note.default_filetype}"
            path = RelPath(RelPath(config.database.file).path.with_name(filename))
        else:
            path = RelPath(path_str)

        return path

    @override
    def execute(self) -> None:
        LOGGER.debug("Starting Note command.")

        Event.PreNoteCommand.fire(self)

        bib = Database()

        try:
            entry = bib[self.largs.label]
        except KeyError:
            # No entry for given label found
            msg = f"No entry with the label '{self.largs.label}' could be found."
            LOGGER.error(msg)
            return

        # NOTE: We hack into the note field and provide an otherwise impossible value in order to
        # detect faulty data in this field from versions prior to the existence of the note command.
        if entry.note is False:  # type: ignore[comparison-overlap]
            LOGGER.error(  # type: ignore[unreachable]
                "The 'note' field of the '%s' entry is faulty! Check the logs during the loading of"
                " the database or the `lint` command for more details.",
                entry.label,
            )
            return

        path = self.note_path(entry)

        task_desc = ""

        if self.largs.action == "edit":
            self.edit(path)
            if not path.path.exists():
                msg = (
                    f"Could not find the note file associated with '{self.largs.label}'. Skipping "
                    "any further actions. Check the following path if you believe this to be an "
                    f"error: '{path}'"
                )
                LOGGER.warning(msg)
                return
            entry.note = str(path)
            bib.update({entry.label: entry})
            bib.save()
            task_desc = "edited"

        elif self.largs.action == "show":
            try:
                self.note_content = open(path.path, "r", encoding="utf-8").read()
            except FileNotFoundError as exc:
                LOGGER.error(
                    "Encountered the following error while trying to read the note of the entry "
                    f"'{self.largs.label}':"
                )
                LOGGER.error(exc)
            task_desc = "shown"

        elif self.largs.action == "delete":
            path.path.unlink(missing_ok=True)
            if path.path.exists():
                # NOTE: it should be impossible to get here unless the .unlink() call above fails
                # for some reason
                msg = (
                    f"Could not delete the note file associated with '{self.largs.label}'. Check "
                    f"the following path: '{path}'"
                )
                LOGGER.warning(msg)
                return
            try:
                entry.note = None
            except KeyError:
                msg = f"The entry '{self.largs.label}' did not have an associated note!"
                LOGGER.warning(msg)
            bib.update({entry.label: entry})
            bib.save()
            task_desc = "deleted"

        elif self.largs.action == "_inline":
            LOGGER.info(
                "An inline note edit was performed. This command is merely executed to ensure it "
                "gets committed into any git history tracking."
            )
            entry.note = str(path)
            bib.update({entry.label: entry})
            bib.save()
            task_desc = "edited"

        else:
            LOGGER.warning(
                f"Encountered unexpected command action: '{self.largs.action}'! "
                "Don't know what to do!"
            )
            task_desc = "ignored"

        Event.PostNoteCommand.fire(self)

        self.git(add_files=[str(path)])

        msg = f"The note of '{self.largs.label}' was successfully {task_desc}."
        LOGGER.info(msg)

    @staticmethod
    def edit(path: RelPath) -> None:
        """Spawns an editor in the terminal to edit the provided file.

        Args:
            path: the file to edit.
        """
        app = get_active_app()
        if app is None:
            NoteCommand._edit(path)
        else:
            with app.suspend():
                NoteCommand._edit(path)

    @staticmethod
    def _edit(path: RelPath) -> None:
        LOGGER.debug('Starting editor "%s".', config.commands.edit.editor)
        status = os.system(config.commands.edit.editor + " " + str(path))
        if status == 0:
            LOGGER.debug("Editor finished successfully.")
        else:
            LOGGER.error(
                f'The note editing using "{config.commands.edit.editor}" failed with the following '
                f"error code: {status}"
            )

    @override
    def render_rich(self) -> ConsoleRenderable | None:
        if self.note_content is None:
            return None
        syntax = Syntax(
            self.note_content,
            config.commands.note.default_filetype,
            theme=config.theme.syntax.get_theme(),
            background_color=config.theme.syntax.get_background_color(),
            line_numbers=config.theme.syntax.line_numbers,
            word_wrap=True,
        )
        return syntax



"""coBib's diff renderer utility."""

from __future__ import annotations

import difflib
import re
from collections import namedtuple
from dataclasses import dataclass, field

from rich.style import Style
from rich.syntax import Syntax
from rich.table import Table

from cobib.config import config

Coordinate = namedtuple("Coordinate", ["row", "col"])


@dataclass
class Side:
    """A simple dataclass for configuring the different sides of a diff view."""

    name: str
    """The name of the side."""
    background: str
    """What background color to use for highlighted differences on this side."""
    foreground: str
    """What foreground color to use for highlighted differences on this side."""


LEFT = Side("a", "#300000", "red")
"""The configured left side of a diff view."""
RIGHT = Side("b", "#002000", "green")
"""The configured right side of a diff view."""


@dataclass
class DiffData:
    """A simple dataclass for holding the difference data."""

    lines: list[str]
    """The total lines being compared."""

    highlight_lines: set[int] = field(default_factory=set)
    """The indices of the lines to be highlighted because they differ."""

    style_ranges: list[tuple[Style, Coordinate, Coordinate]] = field(default_factory=list)
    """A list of stylized ranges to further highlight differences. This is a list of tuples, each of
    which contains three elements:

    1. the style to be applied to this range
    2. the initial `(row, col)` coordinate
    3. the final `(row, col)` coordinate
    """

    current_block: list[str] = field(default_factory=list)
    """The current list of lines being compared."""

    current_offset: int = 0
    """The current index offset of the lines being compared."""

    def _get_cumulative_count(self) -> int:
        """Returns the cumulative number of characters up to the current offset."""
        return sum(len(line) for line in self.lines[: self.current_offset - 1])

    def get_row_col_from_pos(self, position: int) -> Coordinate:
        """Converts a character position into a `(row, col)` coordinate.

        Args:
            position: the absolute character position.

        Returns:
            The `(row, col)` coordinate.
        """
        position += self._get_cumulative_count()
        col, row = -1, 0
        cumulative_count = 0
        for line in self.lines:
            cumulative_count += len(line)
            if cumulative_count > position:
                col = position - (cumulative_count - len(line))
                break
            row += 1
        return Coordinate(row + 1, col)

    def stylize_matches(self, matches: list[difflib.Match], side: Side) -> None:
        """Processes the provided matches and registers the styles in `self`.

        Args:
            matches: the list of difference matches to be highlighted.
            side: the styling data for a specific side.
        """
        pos = 0
        for match in matches:
            row1, col1 = self.get_row_col_from_pos(pos)
            row2, col2 = self.get_row_col_from_pos(getattr(match, side.name))
            pos = getattr(match, side.name) + match.size

            for row in range(row1, row2 + 1 * (match.size > 0)):
                self.highlight_lines.add(row)

                self.style_ranges.insert(
                    0, (Style(bgcolor=side.background), Coordinate(row, 0), Coordinate(row + 1, 0))
                )

            self.style_ranges.append(
                (Style(bgcolor=side.foreground), Coordinate(row1, col1), Coordinate(row2, col2))
            )

    def render(self, lexer: str = "text") -> Syntax:
        """Renders the difference data as a `rich` renderable.

        Args:
            lexer: the syntax lexing to apply.

        Returns:
            A `rich` renderable.
        """
        syntax = Syntax(
            "".join(self.lines),
            lexer,
            theme=config.theme.syntax.get_theme(),
            background_color=config.theme.syntax.get_background_color(),
            highlight_lines=self.highlight_lines,
            line_numbers=True,
            word_wrap=True,
        )

        for style in self.style_ranges:
            syntax.stylize_range(*style)

        return syntax


class Differ:
    """A utility class to compare two strings."""

    _unified_diff_header = re.compile(r"@@ -(\d+),?(\d*) \+(\d+),?(\d*) @@")
    """A regex to detect the headers of a `unified_diff` output."""

    def __init__(self, left: str, right: str) -> None:
        """Initializes a `Differ` instance.

        Args:
            left: the left string.
            right: the right string.
        """
        self.left = DiffData(left.splitlines(keepends=True))
        self.right = DiffData(right.splitlines(keepends=True))

    def _sequence_matcher(self) -> None:
        """A utility method to process the `difflib.SequenceMatcher` results."""
        matches = difflib.SequenceMatcher(
            a="".join(self.left.current_block),
            b="".join(self.right.current_block),
        ).get_matching_blocks()

        self.left.stylize_matches(matches, LEFT)
        self.right.stylize_matches(matches, RIGHT)

        self.left.current_block = []
        self.right.current_block = []

    def _parse_unified_diff_header(self, line: str) -> None:
        """A utility to process a `difflib.unified_diff` header."""
        result = Differ._unified_diff_header.search(line)

        if result is not None:
            offset_left, _, offset_right, _ = result.groups()

            self.left.current_offset = int(offset_left)
            self.right.current_offset = int(offset_right)

    def compute(self) -> None:
        """Computes the difference between the two strings.

        This combines the information gathered from `difflib.unified_diff` and additionally compares
        each block with a `difflib.SequenceMatcher`.
        """
        udiff = difflib.unified_diff(
            self.left.lines,
            self.right.lines,
            n=0,
            lineterm="",
        )

        for line in udiff:
            if line[:2] == "@@":
                self._sequence_matcher()
                self._parse_unified_diff_header(line)

            elif line[:1] == "-" and line != "--- ":
                self.left.current_block.append(line[1:])

            elif line[:1] == "+" and line != "+++ ":
                self.right.current_block.append(line[1:])

        self._sequence_matcher()

    def render(self, lexer: str = "text") -> Table:
        """Renders the compared strings in a side-by-side view.

        Args:
            lexer: the syntax lexing to apply.

        Returns:
            A `rich` renderable.
        """
        table = Table(box=None, show_header=False)
        table.add_row(self.left.render(lexer), self.right.render(lexer))
        return table



"""coBib's wrapper around the `re` and `regex` packages.

Since `regex` is an optional dependency that can be used as a drop-in replacement of the builtin
`re` module, this simple module deals with exposing either one.
"""

from __future__ import annotations

import logging

LOGGER = logging.getLogger(__name__)

HAS_OPTIONAL_REGEX = False

try:
    import regex
except ModuleNotFoundError:
    LOGGER.info(
        "Could not find the `regex` package. Falling back to the builtin `re` module. Certain "
        "functionality may not be available without the optional `regex` dependency installed."
    )
    import re as regex  # type: ignore[no-redef]
else:
    HAS_OPTIONAL_REGEX = True
    LOGGER.info("Found the `regex` package.")

__all__ = ["regex", "HAS_OPTIONAL_REGEX"]



"""coBib's context manager."""

from __future__ import annotations

from contextvars import copy_context

from textual.app import App


def get_active_app() -> App[None] | None:
    """Gets the active textual App (if any).

    Returns:
        The active textual App (or None) based on the current context.
    """
    ctx = copy_context()
    app = [var for var in ctx.values() if isinstance(var, App)]
    if len(app):
        return app[0]
    return None



"""coBib's `entry_points` method.

This is merely a simple wrapper around the builtin `importlib.metadata.entry_points` method in order
to bring the ["selectable"](https://docs.python.org/3/library/importlib.metadata.html#entry-points)
behavior to Python versions lower than 3.10.
While this would theoretically be possible via the `backports.entry_points_selectable` package, that
does not also add the `module` attributes to the `EntryPoint` objects which this method also allows
to filter by.

.. warning::

   Since this is merely a utility to support Python < 3.10, once support for that will be dropped,
   this method will be removed without further notice.
"""

from __future__ import annotations

import sys
from importlib.metadata import EntryPoint
from importlib.metadata import entry_points as _entry_points


def entry_points(filter: str) -> set[tuple[EntryPoint, bool]]:
    """A wrapper of `importlib.metadata.entry_points`.

    This method extracts the group of entry points matching the `filter` by name. It will also
    indicate whether each entry point actually comes directly from the module of the same name.

    .. note::
       This method returns a set (rather than the normal list or tuple) to avoid duplicate entries
       which would sometimes appear in lower Python versions during some initial testing.

    Args:
        filter: the name of the group and module to check for.

    Returns:
        The filtered set of `EntryPoint` objects wrapped into a tuple with the second element
        indicating whether it belongs to the queried module.
    """
    if sys.version_info.minor > 9:  # noqa: PLR2004
        return {
            (cls, cls.module.startswith(filter))  # type: ignore[attr-defined,misc]
            for cls in _entry_points(group=filter)  # type: ignore[call-arg]
        }
    return {(cls, cls.value.startswith(filter)) for cls in _entry_points()[filter]}



"""coBib's git manager."""

from __future__ import annotations

import subprocess
from pathlib import Path


def is_inside_work_tree(path: Path) -> bool:
    """Returns whether the provided path is inside a git worktree and is not being ignored.

    The reason that we also check whether the path is being ignored, is to support the tox-based
    testing environment of coBib which places the temporary testing directories inside coBib's own
    git directory structure (but in an ignored location).

    Args:
        path: the path to check.

    Returns:
        Whether the provided path is inside a git worktree and is not being ignored.
    """
    inside_work_tree = subprocess.run(
        ["git", "-C", path, "rev-parse", "--is-inside-work-tree"],
        check=False,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    if inside_work_tree.returncode != 0 or inside_work_tree.stdout.decode().strip() != "true":
        return False
    path_ignored = subprocess.run(
        ["git", "-C", path, "check-ignore", "--quiet", path],
        check=False,
    )
    return path_ignored.returncode != 0



"""coBib's configuration.

This file contains both, the actual implementation of the `Config` classes, as well as the runtime
`config` object, which gets exposed on the module level as `cobib.config.config`.
Note, that this last link will not point to the correct location in the online documentation due to
the nature of the lower-level import.
"""
# ruff: noqa: E731, RUF009

from __future__ import annotations

import importlib.util
import io
import logging
import os
import sys
from abc import abstractmethod
from collections.abc import Callable
from dataclasses import MISSING, dataclass, field, fields
from enum import Enum, auto
from pathlib import Path
from typing import TYPE_CHECKING, NamedTuple, TextIO

from rich.style import Style
from rich.theme import Theme
from textual.app import DEFAULT_COLORS
from textual.design import ColorSystem
from typing_extensions import override

from cobib.utils.context import get_active_app
from cobib.utils.regex import HAS_OPTIONAL_REGEX
from cobib.utils.rel_path import RelPath

if TYPE_CHECKING:
    import rich

    from .event import Event

LOGGER = logging.getLogger(__name__)
"""@private module logger."""


@dataclass
class _ConfigBase:
    """Base class for configuration section dataclasses."""

    @staticmethod
    def _assert(expression: bool, error: str) -> None:
        """Asserts the expression is True.

        Args:
            expression: the expression to assert.
            error: the message of the `RuntimeError` upon assertion failure.

        Raises:
            RuntimeError with the specified error string.
        """
        if not expression:
            raise RuntimeError(error)

    @abstractmethod
    def validate(self) -> None:
        """Validates the configuration at runtime.

        Raises:
            RuntimeError when an invalid setting is encountered.
        """

    def defaults(self) -> None:
        """Resets the configuration to the default settings."""
        for name, field_ in self.__dataclass_fields__.items():
            if field_.default != MISSING:
                setattr(self, name, field_.default)
            else:
                setattr(self, name, field_.default_factory())  # type: ignore[misc]


class TagMarkup(NamedTuple):
    """A tuple for representing the weight and style of a tag."""

    weight: int
    """The weight of the tag. This integer is used to determine the markup priority of the tag.
    Higher integer values indicate a higher priority."""
    style: str | rich.style.Style
    """The style of the tag. This can be a `rich.style.Style` or a string which can be interpreted
    by `rich.style.Style.parse`."""


@dataclass
class TagsThemeConfig(_ConfigBase):
    """The markup configuration for special tags.

    The tag names configured via this setting will be marked up via a `rich.Theme`. This allows you
    to easily customize a visual differentiation of your entries based on simple properties.

    The style and weight of the builtin special tags can be configured directly or you can add fully
    custom special tags via the `user_tags` field.
    """

    new: TagMarkup = TagMarkup(10, "bold bright_cyan")
    """The markup of the `new` tag. By default, coBib does *not* add this automatically, but you can
    do this easily with a `PostAddCommand` hook like so:

    ```python
    @Event.PostAddCommand.subscribe
    def add_new_tag(cmd: AddCommand) -> None:
        for entry in cmd.new_entries.values():
            if "new" not in entry.tags:
                entry.tags = entry.tags + ["new"]
    ```

    .. note::
       The `new` tag has a lower weight than all of the builtin priority tags (`high`, `medium`,
       `low`) allowing these to be used to further classify new entries on a reading list.
    """
    high: TagMarkup = TagMarkup(40, "on bright_red")
    """The markup of the `high` priority tag."""
    medium: TagMarkup = TagMarkup(30, "bright_red")
    """The markup of the `medium` priority tag."""
    low: TagMarkup = TagMarkup(20, "bright_yellow")
    """The markup of the `low` priority tag."""
    user_tags: dict[str, TagMarkup] = field(default_factory=dict)
    """A dictionary to add weight and markup definitions for arbitrary tags. The keys of the
    dictionary will be compared as is to the tags of an `cobib.database.entry.Entry`.

    .. note::
       Because the markup names are used in a `rich.Theme`, they must be lower case, start with a
       letter, and only contain letters or the characters `"."`, `"-"`, `"_"`. See also
       [here](https://rich.readthedocs.io/en/stable/style.html#style-themes).
    """

    @property
    def names(self) -> set[str]:
        """Returns the set of all special tag names."""
        return {f.name for f in fields(self)} - {"user_tags"} | set(self.user_tags.keys())

    @property
    def styles(self) -> dict[str, str | Style]:
        """Returns the combined dictionary of all special tag styles."""
        styles: dict[str, str | Style] = {}
        for field_ in fields(self):
            tag_markup = getattr(self, field_.name)
            if isinstance(tag_markup, TagMarkup):
                styles[f"tag.{field_.name}"] = tag_markup.style

        for name, weighted_style in self.user_tags.items():
            styles[f"tag.{name}"] = weighted_style.style

        return styles

    @property
    def weights(self) -> dict[str, int]:
        """Returns the combined dictionary of all special tag weights."""
        weights: dict[str, int] = {}
        for field_ in fields(self):
            tag_markup = getattr(self, field_.name)
            if isinstance(tag_markup, TagMarkup):
                weights[field_.name] = tag_markup.weight

        for name, weighted_style in self.user_tags.items():
            weights[name] = weighted_style.weight

        return weights

    @override
    def validate(self) -> None:
        self._assert(
            isinstance(self.new, TagMarkup), "config.theme.tags.new should be a TagMarkup tuple."
        )
        self._assert(
            isinstance(self.high, TagMarkup), "config.theme.tags.high should be a TagMarkup tuple."
        )
        self._assert(
            isinstance(self.medium, TagMarkup),
            "config.theme.tags.medium should be a TagMarkup tuple.",
        )
        self._assert(
            isinstance(self.low, TagMarkup), "config.theme.tags.low should be a TagMarkup tuple."
        )
        self._assert(
            isinstance(self.user_tags, dict),
            "config.theme.tags.user_tags should be a dict.",
        )
        for name, tag_markup in self.user_tags.items():
            self._assert(
                isinstance(tag_markup, TagMarkup),
                f"The '{name}' entry in config.theme.tags.user_tags should be a TagMarkup tuple.",
            )


@dataclass
class SearchHighlightConfig(_ConfigBase):
    """The `config.theme.search` section."""

    label: str | Style = "blue"
    """Specifies the color with which to highlight the labels of search results."""
    query: str | Style = "red"
    """Specifies the color with which to highlight the query matches of a search."""

    @property
    def styles(self) -> dict[str, str | Style]:
        """Returns the `rich.Theme`-compatible styles dictionary of the configured highlights."""
        return {
            "search.label": self.label,
            "search.query": self.query,
        }

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the THEME.SEARCH configuration section.")
        self._assert(isinstance(self.label, str), "config.theme.search.label should be a string.")
        self._assert(isinstance(self.query, str), "config.theme.search.query should be a string.")


@dataclass
class SyntaxConfig(_ConfigBase):
    """The `config.theme.syntax` section.

    These attributes configure the default values used for displaying `rich.Syntax` elements.
    """

    theme: str | None = None
    """The theme to use for any `rich.Syntax` display.

    If this is `None`, it defaults to `"ansi_dark"` or `"ansi_light"`, depending on the value of
    `cobib.config.ThemeConfig.dark`.

    Otherwise, this should be the name of a supported pygments theme. For more details refer to
    [rich's documentation](https://rich.readthedocs.io/en/latest/syntax.html#theme).

    This attribute should be accessed via the `SyntaxConfig.get_theme` method.
    """

    background_color: str | None = None
    """The background color to use for any `rich.Syntax` display.

    If this is `None`, its default value depends on whether a `textual.App` is running. If not, it
    simply uses `"default"` which should result in a _transparent_ background with respect to the
    terminal's background color. Otherwise, it will use the `$panel` color of the `textual.App`. For
    more details refer to [textual's
    documentation](https://textual.textualize.io/guide/design/#base-colors).

    This attribute should be accessed via the `SyntaxConfig.get_background_color` method.
    """

    line_numbers: bool = True
    """Whether to show line numbers in `rich.Syntax` displays.

    .. note::
       This setting is ignored by the `cobib.utils.diff_renderer.DiffData.render` method. It will
       *always* show line numbers.
    """

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the THEME.SYNTAX configuration section.")
        self._assert(
            self.theme is None or isinstance(self.theme, str),
            "config.theme.syntax.theme should either be `None` or a string.",
        )
        self._assert(
            self.background_color is None or isinstance(self.background_color, str),
            "config.theme.syntax.background_color should either be `None` or a string.",
        )
        self._assert(
            isinstance(self.line_numbers, bool),
            "config.theme.syntax.line_numbers should be a boolean.",
        )

    def get_theme(self) -> str:
        """Returns the `SyntaxConfig.theme` value."""
        if self.theme is not None:
            return self.theme

        return "ansi_dark" if config.theme.dark else "ansi_light"

    def get_background_color(self) -> str:
        """Returns the `SyntaxConfig.background_color` value."""
        if self.background_color is not None:
            return self.background_color

        if get_active_app() is None:
            return "default"

        return config.theme.css_variables["panel"]


@dataclass
class ThemeConfig(_ConfigBase):
    """The `config.theme` section."""

    dark: bool = True
    """Whether a dark or light theme should be used."""
    design: dict[str, ColorSystem] = field(default_factory=lambda: DEFAULT_COLORS)
    """A dictionary mapping theme names to `textual.design.ColorSystem`s.

    Here is a simple example how you can add an intense splash of color to the default ones:

    ```python
    from textual.app import DEFAULT_COLORS
    from textual.color import Color

    a_splash_of_pink = DEFAULT_COLORS.copy()
    a_splash_of_pink["dark"].primary = Color.parse("#ff00ff")
    a_splash_of_pink["light"].primary = Color.parse("#ff00ff")
    config.theme.design = a_splash_of_pink
    ```

    Refer to [textual's documentation](https://textual.textualize.io/guide/design) for more details.
    """
    search: SearchHighlightConfig = field(default_factory=lambda: SearchHighlightConfig())
    """The nested section for theme settings related to the `search` command."""
    syntax: SyntaxConfig = field(default_factory=lambda: SyntaxConfig())
    """The nested section for theme settings related to `rich.Syntax` displays."""
    tags: TagsThemeConfig = field(default_factory=lambda: TagsThemeConfig())
    """The nested section for the markup of special tags."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the THEME configuration section.")
        self._assert(isinstance(self.dark, bool), "config.theme.dark should be a boolean.")
        self._assert(isinstance(self.design, dict), "config.theme.design should be a dictionary.")
        for name, colors in self.design.items():
            self._assert(
                isinstance(colors, ColorSystem),
                (
                    f"The '{name}' entry in config.theme.design should be a "
                    "`textual.design.ColorSystem` instance."
                ),
            )
        self.search.validate()
        self.syntax.validate()
        self.tags.validate()

    def build(self) -> Theme:
        """Returns the built `rich.Theme` from the configured styles."""
        theme: dict[str, str | Style] = {}
        theme.update(self.search.styles)
        theme.update(self.tags.styles)
        return Theme(theme)

    @property
    def css_variables(self) -> dict[str, str]:
        """The actual CSS color variables generated from the active theme of `ThemeConfig.design`.

        The active theme is either `"dark"` or `"light"` depending on `ThemeConfig.dark`.

        Returns:
            A dictionary mapping from color names to values. See [textual's
            documentation](https://textual.textualize.io/guide/design) for more details.
        """
        return self.design["dark" if self.dark else "light"].generate()


@dataclass
class LoggingConfig(_ConfigBase):
    """The `config.logging` section."""

    cache: str | Path = "~/.cache/cobib/cache"
    """Specifies the default cache location."""
    logfile: str | Path = "~/.cache/cobib/cobib.log"
    """Specifies the default logfile location."""
    version: str | None = "~/.cache/cobib/version"
    """Specifies the location for the cached version number based on which coBib shows the latest
    changes. Set this to `None` to disable this functionality entirely."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the LOGGING configuration section.")
        self._assert(
            isinstance(self.cache, (str, Path)), "config.logging.cache should be a string."
        )
        self._assert(
            isinstance(self.logfile, (str, Path)), "config.logging.logfile should be a string."
        )
        self._assert(
            self.version is None or isinstance(self.version, str),
            "config.logging.version should be a string or `None`.",
        )


@dataclass
class AddCommandConfig(_ConfigBase):
    """The `config.commands.add` section."""

    skip_download: bool = False
    """Specifies whether to skip the attempt of downloading PDF files of added entries."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the COMMANDS.ADD configuration section.")
        self._assert(
            isinstance(self.skip_download, bool),
            "config.commands.add.skip_download should be a boolean.",
        )


@dataclass
class DeleteCommandConfig(_ConfigBase):
    """The `config.commands.delete` section."""

    confirm: bool = True
    """Whether or not to confirm before deleting an entry."""

    preserve_files: bool = False
    """Specifies whether associated files should be preserved when deleting an entry."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the COMMANDS.DELETE configuration section.")
        self._assert(
            isinstance(self.confirm, bool),
            "config.commands.delete.confirm should be a boolean.",
        )
        self._assert(
            isinstance(self.preserve_files, bool),
            "config.commands.delete.preserve_files should be a boolean.",
        )


@dataclass
class EditCommandConfig(_ConfigBase):
    """The `config.commands.edit` section."""

    default_entry_type: str = "article"
    """Specifies the default bibtex entry type."""
    editor: str = os.environ.get("EDITOR", "vim")
    """Specifies the editor program. Note, that this default will respect your `$EDITOR`
    environment setting and fall back to `vim` if that variable is not set."""
    preserve_files: bool = False
    """Specifies whether associated files should be preserved when renaming an entry during
    editing."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the COMMANDS.EDIT configuration section.")
        self._assert(
            isinstance(self.default_entry_type, str),
            "config.commands.edit.default_entry_type should be a string.",
        )
        self._assert(
            isinstance(self.editor, str),
            "config.commands.edit.editor should be a string.",
        )
        self._assert(
            isinstance(self.preserve_files, bool),
            "config.commands.edit.preserve_files should be a boolean.",
        )


@dataclass
class ImportCommandConfig(_ConfigBase):
    """The `config.commands.import` section."""

    skip_download: bool = False
    """Specifies whether to skip the attempt of downloading PDF files of imported entries."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the COMMANDS.IMPORT configuration section.")
        self._assert(
            isinstance(self.skip_download, bool),
            "config.commands.import.skip_download should be a boolean.",
        )


@dataclass
class ListCommandConfig(_ConfigBase):
    """The `config.commands.list_` section."""

    default_columns: list[str] = field(default_factory=lambda: ["label", "title"])
    """Specifies the default columns shown during the `list` command."""
    ignore_case: bool = False
    """Specifies whether filter matching should be performed case-insensitive."""
    decode_unicode: bool = False
    """Specifies whether filter matching should decode all Unicode characters."""
    decode_latex: bool = False
    """Specifies whether filter matching should decode all LaTeX sequences."""
    fuzziness: int = 0
    """The amount of fuzzy errors to allow for filter matching. Using this feature requires the
    optional `regex` dependency to be installed."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the COMMANDS.LIST configuration section.")
        self._assert(
            isinstance(self.default_columns, list),
            "config.commands.list_.default_columns should be a list.",
        )
        self._assert(
            isinstance(self.ignore_case, bool),
            "config.commands.list_.ignore_case should be a boolean.",
        )
        self._assert(
            isinstance(self.decode_unicode, bool),
            "config.commands.list_.decode_unicode should be a boolean.",
        )
        self._assert(
            isinstance(self.decode_latex, bool),
            "config.commands.list_.decode_latex should be a boolean.",
        )
        self._assert(
            isinstance(self.fuzziness, int) and self.fuzziness >= 0,
            "config.commands.list_.fuzziness should be a non-negative integer.",
        )
        if self.fuzziness > 0 and not HAS_OPTIONAL_REGEX:  # pragma: no branch
            LOGGER.warning(  # pragma: no cover
                "Using `config.commands.list_.fuzziness` requires the optional `regex` "
                "dependency to be installed! Falling back to `fuzziness=0`."
            )
            self.fuzziness = 0  # pragma: no cover


@dataclass
class ModifyCommandConfig(_ConfigBase):
    """The `config.commands.modify` section."""

    preserve_files: bool = False
    """Specifies whether associated files should be preserved when renaming an entry during
    modifying."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the COMMANDS.MODIFY configuration section.")
        self._assert(
            isinstance(self.preserve_files, bool),
            "config.commands.modify.preserve_files should be a boolean.",
        )


@dataclass
class NoteCommandConfig(_ConfigBase):
    """The `config.commands.note` section."""

    default_filetype: str = "txt"
    """Specifies the default filetype for notes."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the COMMANDS.NOTE configuration section.")
        self._assert(
            isinstance(self.default_filetype, str),
            "config.commands.note.default_filetype should be a string.",
        )


@dataclass
class OpenCommandConfig(_ConfigBase):
    """The `config.commands.open` section."""

    command: str = "xdg-open" if sys.platform.lower() == "linux" else "open"
    """Specifies the command used for opening files associated with your entries."""
    fields: list[str] = field(default_factory=lambda: ["file", "url"])
    """Specifies the entry fields which are to be checked for openable URLs."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the COMMANDS.OPEN configuration section.")
        self._assert(
            isinstance(self.command, str),
            "config.commands.open.command should be a string.",
        )
        self._assert(
            isinstance(self.fields, list),
            "config.commands.open.fields should be a list.",
        )


@dataclass
class SearchCommandConfig(_ConfigBase):
    """The `config.commands.search` section."""

    context: int = 1
    """Specifies the default number of context line to provide for each search query match. This is
    similar to the `-C` option of `grep`."""
    grep: str = "grep"
    """Specifies the grep tool used for searching through your database and associated files. The
    default tool (`grep`) will not provide results for attached PDFs but other tools such as
    [ripgrep-all](https://github.com/phiresky/ripgrep-all) will."""
    grep_args: list[str] = field(default_factory=list)
    """Specifies additional arguments for your grep command. Note, that GNU's grep understands
    extended regex patterns even without specifying `-E`."""
    ignore_case: bool = False
    """Specifies whether searches should be performed case-insensitive."""
    decode_unicode: bool = False
    """Specifies whether searches should decode all Unicode characters."""
    decode_latex: bool = False
    """Specifies whether searches should decode all LaTeX sequences."""
    fuzziness: int = 0
    """The amount of fuzzy errors to allow for search matches. Using this feature requires the
    optional `regex` dependency to be installed."""
    skip_files: bool = False
    """Specifies whether searches should skip looking through associated files."""
    skip_notes: bool = False
    """Specifies whether searches should skip looking through associated notes."""

    @property
    def highlights(self) -> SearchHighlightConfig:
        """**DEPRECATED** Use `config.theme.search` instead!

        The nested section for highlights used when displaying search results.
        """
        LOGGER.log(
            45,
            "The config.commands.search.highlights setting is DEPRECATED! Please use the new "
            "config.theme.search setting instead.",
        )
        return config.theme.search

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the COMMANDS.SEARCH configuration section.")
        self._assert(
            isinstance(self.context, int) and self.context >= 0,
            "config.commands.search.context should be a non-negative integer.",
        )
        self._assert(
            isinstance(self.grep, str),
            "config.commands.search.grep should be a string.",
        )
        self._assert(
            isinstance(self.grep_args, list),
            "config.commands.search.grep_args should be a list.",
        )
        self._assert(
            isinstance(self.ignore_case, bool),
            "config.commands.search.ignore_case should be a boolean.",
        )
        self._assert(
            isinstance(self.decode_unicode, bool),
            "config.commands.search.decode_unicode should be a boolean.",
        )
        self._assert(
            isinstance(self.decode_latex, bool),
            "config.commands.search.decode_latex should be a boolean.",
        )
        self._assert(
            isinstance(self.fuzziness, int) and self.fuzziness >= 0,
            "config.commands.search.fuzziness should be a non-negative integer.",
        )
        if self.fuzziness > 0 and not HAS_OPTIONAL_REGEX:
            LOGGER.warning(
                "Using `config.commands.search.fuzziness` requires the optional `regex` "
                "dependency to be installed! Falling back to `fuzziness=0`."
            )
            self.fuzziness = 0
        self._assert(
            isinstance(self.skip_files, bool),
            "config.commands.search.skip_files should be a boolean.",
        )
        self._assert(
            isinstance(self.skip_notes, bool),
            "config.commands.search.skip_notes should be a boolean.",
        )


@dataclass
class ShowCommandConfig(_ConfigBase):
    """The `config.commands.show` section."""

    encode_latex: bool = True
    """Specifies whether non-ASCII characters should be encoded using LaTeX sequences."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the COMMANDS.SHOW configuration section.")
        self._assert(
            isinstance(self.encode_latex, bool),
            "config.commands.show.encode_latex should be a boolean.",
        )


@dataclass
class CommandConfig(_ConfigBase):
    """The `config.commands` section."""

    add: AddCommandConfig = field(default_factory=lambda: AddCommandConfig())
    """The nested section for settings related to the `add` command."""
    delete: DeleteCommandConfig = field(default_factory=lambda: DeleteCommandConfig())
    """The nested section for settings related to the `delete` command."""
    edit: EditCommandConfig = field(default_factory=lambda: EditCommandConfig())
    """The nested section for settings related to the `edit` command."""
    import_: ImportCommandConfig = field(default_factory=lambda: ImportCommandConfig())
    """The nested section for settings related to the `import` command. Note the trailing underscore
    of its name, since this attribute would otherwise clash with the builtin `import` keyword."""
    list_: ListCommandConfig = field(default_factory=lambda: ListCommandConfig())
    """The nested section for settings related to the `list` command. Note the trailing underscore
    of its name, since this attribute would otherwise clash with the builtin `list` keyword."""
    modify: ModifyCommandConfig = field(default_factory=lambda: ModifyCommandConfig())
    """The nested section for settings related to the `modify` command."""
    note: NoteCommandConfig = field(default_factory=lambda: NoteCommandConfig())
    """The nested section for settings related to the `note` command."""
    open: OpenCommandConfig = field(default_factory=lambda: OpenCommandConfig())
    """The nested section for settings related to the `open` command."""
    search: SearchCommandConfig = field(default_factory=lambda: SearchCommandConfig())
    """The nested section for settings related to the `search` command."""
    show: ShowCommandConfig = field(default_factory=lambda: ShowCommandConfig())
    """The nested section for settings related to the `show` command."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the COMMANDS configuration section.")
        self.add.validate()
        self.delete.validate()
        self.edit.validate()
        self.import_.validate()
        self.list_.validate()
        self.modify.validate()
        self.note.validate()
        self.open.validate()
        self.search.validate()
        self.show.validate()


class LabelSuffix(Enum):
    """Suffixes to disambiguate `cobib.database.Entry` labels."""

    ALPHA = lambda count: chr(96 + count)
    """Enumerates with lowercase roman letters: `a-z`."""
    CAPITAL = lambda count: chr(64 + count)
    """Enumerates with uppercase roman letters: `A-Z`."""

    NUMERIC = lambda count: str(count)
    """Enumerates with arabic numbers: `1, 2, ...`."""

    @staticmethod
    def reverse(type_: LabelSuffix, suffix: str) -> int:
        """Reverses a label suffix enumerator from a suffix to its integer value.

        Args:
            type_: the `LabelSuffix` enumerator.
            suffix: the suffix to reverse.

        Returns:
            The integer value of the reversed suffix.

        Raises:
            ValueError: if a non-numeric suffix is longer than 1 character.
            ValueError: if an alphabetic suffix did not fall within the alphabet.
            ValueError: if an invalid suffix of some other form is provided.
        """
        # TODO: once Python 3.10 becomes the default, change this to a match statement

        if type_ == LabelSuffix.NUMERIC:
            return int(suffix)

        if len(suffix) > 1:
            raise ValueError("A non-numeric suffix may not be longer than a single character.")

        if type_ in {LabelSuffix.ALPHA, LabelSuffix.CAPITAL}:
            # we can convert both of these cases in the same way by using base 36 for `int()`
            ret = int(suffix, 36) - 9
            if ret < 0:
                raise ValueError(
                    f"'{suffix}' is not a valid alphabetic suffix because its ANSI value falls "
                    "outside of the alphabetic range, yielding a negative value."
                )
            return ret

        raise ValueError(f"'{suffix}' is not a valid suffix.")

    @staticmethod
    def suffix_type(suffix: str) -> LabelSuffix | None:
        """Determines the enumerator type for a given suffix.

        Args:
            suffix: the suffix whose enumerator type to determine.

        Returns:
            An optional enumerator type. If the suffix was invalid, `None` will be returned.
        """
        if not suffix.isascii():
            # if the input contains a non-ASCII character, it cannot be a label suffix
            return None

        try:
            # test if the suffix is numeric by attempting to convert it to a base-10 integer
            _ = int(suffix)
        except ValueError:
            # first ensure that we have only alphabetic characters left
            if not suffix.isalpha():
                # this is important, because for example "hello123" will pass `islower()` because
                # all of the cased characters are indeed lower-cased
                return None

            if suffix.islower():
                return LabelSuffix.ALPHA
            if suffix.isupper():
                return LabelSuffix.CAPITAL
        else:
            return LabelSuffix.NUMERIC

        # the provided suffix did not match any of the builtin enumerators
        return None

    @staticmethod
    def trim_label(label: str, separator: str, type_: LabelSuffix) -> tuple[str, int]:
        """Trims the provided label based on the separator and suffix type.

        Args:
            label: the label whose suffix to trim.
            separator: the separator character between the actual label and the suffix.
            type_: the `LabelSuffix` enumerator.

        Returns:
            The pair of the trimmed label and numeric value of its original suffix.
        """
        # initialize the counter of the suffix
        suffix_value = 0

        # try split the suffix from the label using the provided separator
        if separator:
            *pieces, suffix = label.split(separator)
        else:
            # no character is used for separation, we simply fall back to using the last one
            suffix = label[-1]
            pieces = [label[:-1]]

        if pieces:
            # piece together the left-over raw label without its suffix
            raw_label = separator.join(pieces)
            # determine the suffix type of this label
            suffix_type = LabelSuffix.suffix_type(suffix)

            if suffix_type is not None and suffix_type != type_:
                # if it does not match the expected suffix type, ignore it
                LOGGER.info(
                    f"The suffix type encountered on '{label}' does not match the configured one. "
                    "Assuming that this is intentional and therefore ignoring this as a suffix."
                )
                raw_label = label
            else:
                try:
                    # try to obtain the suffix value by reversing it
                    suffix_value = LabelSuffix.reverse(type_, suffix)
                except ValueError:
                    # if this fails, reset the raw label and assume that this was not an actual
                    # disambiguation suffix
                    raw_label = label
        else:
            # the original split did not yield any pieces so the `suffix` is the actual `raw_label`
            raw_label = suffix

        return (raw_label, suffix_value)


class AuthorFormat(Enum):
    """Storage formats for the `author` information."""

    YAML = auto()
    """Stores the list of authors as a YAML list, separating out the first and last names as well as
    name pre- and suffixes."""
    BIBLATEX = auto()
    """Stores the author in the same form as it would be encoded inside of BibLaTeX."""


@dataclass
class DatabaseFormatConfig(_ConfigBase):
    """The `config.database.format` section."""

    author_format: AuthorFormat = AuthorFormat.YAML
    """Specifies the format to use for the `author` information. See also `AuthorFormat` for more
    details."""
    label_default: str = "{unidecode(label)}"
    """Specifies a default label format which will be used for database entry keys. The format of
    this option follows the f-string like formatting of modifications (see also the documentation
    of the `cobib.commands.modify.ModifyCommand`). The default configuration value passes the
    originally provided label through [text-unidecode](https://pypi.org/project/text-unidecode/)
    which replaces all Unicode symbols with pure ASCII ones. A more useful example is
    `"{unidecode(author[0].last)}{year}"` which takes the surname of the first author, replaces the
    Unicode characters and then immediately appends the publication year."""
    label_suffix: tuple[str, LabelSuffix] = field(default_factory=lambda: ("_", LabelSuffix.ALPHA))
    """Specifies the suffix format which is used to disambiguate labels if a conflict would occur.
    This option takes a tuple of length 2, where the first entry is the string separating the
    proposed label from the enumerator and the second one is one of the enumerators provided by the
    `LabelSuffix` object. The available enumerators are:
        - ALPHA: a, b, ...
        - CAPITAL: A, B, ...
        - NUMERIC: 1, 2, ...
    """
    suppress_latex_warnings: bool = True
    """Specifies whether latex warnings should not be ignored during the escaping of special
    characters. This is a simple option which gets passed on to the internally used `pylatexenc`
    library."""
    verbatim_fields: list[str] = field(default_factory=lambda: ["file", "url"])
    """Specifies the fields which will be left verbatim and, thus, remain unaffected from any
    special character conversions (e.g. LaTeX encoding)."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the DATABASE.FORMAT configuration section.")
        self._assert(
            isinstance(self.author_format, AuthorFormat),
            "config.database.format.author_format should be an AuthorFormat value.",
        )
        self._assert(
            isinstance(self.label_default, str),
            "config.database.format.label_default should be a string.",
        )
        self._assert(
            isinstance(self.label_suffix, tuple) and len(self.label_suffix) == 2,  # noqa: PLR2004
            "config.database.format.label_suffix should be a tuple of length 2.",
        )
        self._assert(
            isinstance(self.label_suffix[0], str),
            "The first entry of config.database.format.label_suffix should be a string.",
        )
        self._assert(
            callable(self.label_suffix[1]),
            "The first entry of config.database.format.label_suffix should be a function.",
        )
        self._assert(
            isinstance(self.suppress_latex_warnings, bool),
            "config.database.format.suppress_latex_warnings should be a boolean.",
        )
        self._assert(
            isinstance(self.verbatim_fields, list),
            "config.database.format.verbatim_fields should be a list.",
        )


@dataclass
class EntryListSeparatorConfig(_ConfigBase):
    """The `config.database.stringify.list_separator` section.

    These settings configure how list-based `cobib.database.Entry` fields are transformed into
    strings when converting to the BibTeX format. Each of these fields will be joined by the
    respective values.
    """

    file: str = ", "
    """Specifies the string used to join the list of files into a single string representation."""
    tags: str = ", "
    """Specifies the string used to join the list of tags into a single string representation."""
    url: str = ", "
    """Specifies the string used to join the list of URLs into a single string representation."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the DATABASE.STRINGIFY.LIST_SEPARATOR configuration section.")
        self._assert(
            isinstance(self.file, str),
            "config.database.stringify.list_separator.file should be a string.",
        )
        self._assert(
            isinstance(self.tags, str),
            "config.database.stringify.list_separator.tags should be a string.",
        )
        self._assert(
            isinstance(self.url, str),
            "config.database.stringify.list_separator.url should be a string.",
        )


@dataclass
class EntryStringifyConfig(_ConfigBase):
    """The `config.database.stringify` section."""

    list_separator: EntryListSeparatorConfig = field(
        default_factory=lambda: EntryListSeparatorConfig()
    )
    """The nested section for list separator values."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the DATABASE.STRINGIFY configuration section.")
        self.list_separator.validate()


@dataclass
class DatabaseConfig(_ConfigBase):
    """The `config.database` section."""

    file: str | Path = "~/.local/share/cobib/literature.yaml"
    """Specifies the path to the database YAML file. You can use `~` to represent your `$HOME`
    directory."""
    cache: str | Path | None = "~/.cache/cobib/databases/"
    """Specifies the folder in which already parsed databases should be stored. Set this to `None`
    if you want to disable caching entirely."""
    format: DatabaseFormatConfig = field(default_factory=lambda: DatabaseFormatConfig())
    """The nested section for database formatting settings."""
    git: bool = False
    """coBib can integrate with `git` in order to automatically track the history of the database.
    However, by default, this option is disabled. In order to make use of this, enable this setting
    and initialize your database with `cobib init --git`.

    .. warning::
       Before enabling this setting you must ensure that you have set up git properly by setting
       your name and email address."""
    stringify: EntryStringifyConfig = field(default_factory=lambda: EntryStringifyConfig())
    """The nested section for database string-formatting settings."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the DATABASE configuration section.")
        self._assert(isinstance(self.file, str), "config.database.file should be a string.")
        self._assert(
            self.cache is None or isinstance(self.cache, str),
            "config.database.cache should be a string or `None`.",
        )
        self._assert(isinstance(self.git, bool), "config.database.git should be a boolean.")
        self.format.validate()
        self.stringify.validate()


@dataclass
class BibtexParserConfig(_ConfigBase):
    """The `config.parsers.bibtex` section."""

    ignore_non_standard_types: bool = False
    """Specifies whether the BibTeX-parser should ignore non-standard BibTeX entry types."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the PARSERS.BIBTEX configuration section.")
        self._assert(
            isinstance(self.ignore_non_standard_types, bool),
            "config.parsers.bibtex.ignore_non_standard_types should be a boolean.",
        )


@dataclass
class YAMLParserConfig(_ConfigBase):
    """The `config.parsers.yaml` section."""

    use_c_lib_yaml: bool = True
    """Specifies whether the C-based implementation of the YAML parser (called `LibYAML`) shall be
    used, *significantly* increasing the performance of the parsing.

    .. note::
       This requires manual installation of the C-based parser:
       https://yaml.readthedocs.io/en/latest/install.html#optional-requirements
    """

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the PARSERS.YAML configuration section.")
        self._assert(
            isinstance(self.use_c_lib_yaml, bool),
            "config.parsers.yaml.use_c_lib_yaml should be a boolean.",
        )


@dataclass
class ParserConfig(_ConfigBase):
    """The `config.parsers` section."""

    bibtex: BibtexParserConfig = field(default_factory=lambda: BibtexParserConfig())
    """The nested section for the BibTeX parser settings."""
    yaml: YAMLParserConfig = field(default_factory=lambda: YAMLParserConfig())
    """The nested section for the YAML parser settings."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the PARSERS configuration section.")
        self.bibtex.validate()
        self.yaml.validate()


@dataclass
class TUIConfig(_ConfigBase):
    """The `config.tui` section."""

    scroll_offset: int = 2
    """The minimum number of lines to keep above and below the cursor in the
    `cobib.ui.tui.components.ListView`. This is similar to Vim's `scrolloff` setting."""
    tree_folding: tuple[bool, bool] = (True, False)
    """The default folding level of the tree nodes in the `cobib.ui.tui.components.SearchView`. When
    `True`, the node will be initialized to be folded (or collapsed), when `False`, it will be
    expanded. The first boolean corresponds to the nodes for each matching entry, the second one is
    for all the search matches."""
    preset_filters: list[str] = field(default_factory=list)
    """Permits providing a list of preset filters. These can be interactively selected in the TUI by
    pressing `p`. To specify these, simply provide a string with the filter arguments, for example:

    ```python
    config.tui.preset_filters = [
        "++tags READING",
        "++year 2023",
    ]
    ```

    The first 9 filters can be quickly accessed in the TUI by simply pressing the corresponding
    number. You can also use 0 to reset any applied filter."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the TUI configuration section.")
        self._assert(
            isinstance(self.scroll_offset, int),
            "config.tui.scroll_offset should be an integer.",
        )
        self._assert(
            isinstance(self.preset_filters, list),
            "config.tui.preset_filters should be a list.",
        )
        for preset in self.preset_filters:
            self._assert(
                isinstance(preset, str),
                "config.tui.preset_filters should be a list of strings.",
            )


@dataclass
class FileDownloaderConfig(_ConfigBase):
    """The `config.utils.file_downloader` section."""

    default_location: str = "~/.local/share/cobib"
    """Specifies the default download location for associated files."""
    url_map: dict[str, str] = field(default_factory=dict)
    """Permits providing rules to map from a journal's landing page URL to its PDF URL. To do so,
    insert an entry into this dictionary, with a regex-pattern matching the journal's landing page
    URL and a value being the PDF URL. E.g.:

    ```python
    config.utils.file_downloader.url_map[
        r"(.+)://aip.scitation.org/doi/([^/]+)"
    ] = r"\1://aip.scitation.org/doi/pdf/\2"

    config.utils.file_downloader.url_map[
        r"(.+)://quantum-journal.org/papers/([^/]+)"
    ] = r"\1://quantum-journal.org/papers/\2/pdf/"
    ```

    Make sure to use raw Python strings to ensure proper backslash-escaping.

    You can find some examples on
    [this wiki page](https://gitlab.com/cobib/cobib/-/wikis/File-Downloader-URL-Maps)."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the UTILS.FILE_DOWNLOADER configuration section.")
        self._assert(
            isinstance(self.default_location, str),
            "config.utils.file_downloader.default_location should be a string.",
        )
        self._assert(
            isinstance(self.url_map, dict),
            "config.utils.file_downloader.url_map should be a dict.",
        )
        for pattern, repl in self.url_map.items():
            self._assert(
                isinstance(pattern, str) and isinstance(repl, str),
                "config.utils.file_downloader.url_map should be a dict[str, str].",
            )


@dataclass
class UtilsConfig(_ConfigBase):
    """The `config.utils` section."""

    file_downloader: FileDownloaderConfig = field(default_factory=lambda: FileDownloaderConfig())
    """The nested section for the `cobib.utils.FileDownloader` utils settings."""
    journal_abbreviations: list[tuple[str, str]] = field(default_factory=list)
    """Permits providing a list of journal abbreviations. This list should be formatted as tuples of
    the form: `(full journal name, abbreviation)`. The abbreviation should include any necessary
    punctuation which can be excluded upon export (see also `cobib export --help`).

    You can find some examples on
    [this wiki page](https://gitlab.com/cobib/cobib/-/wikis/Journal-Abbreviations)."""

    @override
    def validate(self) -> None:
        LOGGER.debug("Validating the UTILS configuration section.")
        self.file_downloader.validate()
        self._assert(
            isinstance(self.journal_abbreviations, list),
            "config.utils.journal_abbreviations should be a list.",
        )
        for abbrev in self.journal_abbreviations:
            self._assert(
                isinstance(abbrev, tuple),
                "config.utils.journal_abbreviations should be a list of tuples.",
            )


@dataclass
class Config(_ConfigBase):
    """The `config` dataclass."""

    logging: LoggingConfig = field(default_factory=lambda: LoggingConfig())
    """The nested section for the logging settings."""
    commands: CommandConfig = field(default_factory=lambda: CommandConfig())
    """The nested section for the commands settings."""
    database: DatabaseConfig = field(default_factory=lambda: DatabaseConfig())
    """The nested section for the database settings."""
    events: dict["Event", list[Callable]] = field(default_factory=dict)  # type: ignore[type-arg]
    """It is possible to register hooks on various events. Although this can be done manually using
    this dictionary, it is preferred to use the function-decorators like so:
    To subscribe to a certain event do something similar to the following:

    ```python
    from os import system
    from cobib.config import Event
    from cobib.commands import InitCommand

    @Event.PostInitCommand.subscribe
    def add_remote(cmd: InitCommand) -> None:
        system(f"git -C {cmd.root} remote add origin https://github.com/user/repo")
    ```

    Note, that the typing is required for the config validation to pass!
    For more information refer to the `cobib.config.Event` documentation.
    """
    parsers: ParserConfig = field(default_factory=lambda: ParserConfig())
    """The nested section for the parsers settings."""
    theme: ThemeConfig = field(default_factory=lambda: ThemeConfig())
    """The nested section for the theme settings."""
    tui: TUIConfig = field(default_factory=lambda: TUIConfig())
    """The nested section for the TUI settings."""
    utils: UtilsConfig = field(default_factory=lambda: UtilsConfig())
    """The nested section for the utils settings."""

    XDG_CONFIG_FILE: str | Path = field(
        default="~/.config/cobib/config.py", init=False, repr=False, compare=False
    )
    """The XDG-based standard configuration location."""

    @override
    def validate(self) -> None:
        LOGGER.info("Validating the runtime configuration.")
        self.logging.validate()
        self.commands.validate()
        self.database.validate()
        self.parsers.validate()
        self.theme.validate()
        self.tui.validate()
        self.utils.validate()

        LOGGER.debug("Validating the EVENTS configuration section.")
        self._assert(isinstance(self.events, dict), "config.events should be a dict.")
        for event in self.events:
            self._assert(
                event.validate(),
                f"config.events.{event} did not pass its validation check.",
            )

    @staticmethod
    def load(configpath: str | Path | TextIO | io.TextIOWrapper | None = None) -> None:
        """Loads another configuration object at runtime.

        WARNING: The new Python-like configuration allows essentially arbitrary Python code so it is
        the user's responsibility to treat this with care!

        Args:
            configpath: the path to the configuration.
        """
        LOGGER.info(configpath)
        if configpath is not None:
            if isinstance(configpath, (TextIO, io.TextIOWrapper)):
                configpath.close()
                configpath = configpath.name
        elif "COBIB_CONFIG" in os.environ:
            configpath_env = os.environ["COBIB_CONFIG"]
            if configpath_env.lower() in ("", "0", "f", "false", "nil", "none"):
                LOGGER.info(
                    "Skipping configuration loading because negative COBIB_CONFIG environment "
                    "variable was detected."
                )
                return
            configpath = RelPath(configpath_env).path
        elif Config.XDG_CONFIG_FILE and RelPath(Config.XDG_CONFIG_FILE).exists():
            configpath = RelPath(Config.XDG_CONFIG_FILE).path
        else:  # pragma: no cover
            return  # pragma: no cover
        LOGGER.info("Loading configuration from default location: %s", configpath)

        spec = importlib.util.spec_from_file_location("config", configpath)
        if spec is None:
            LOGGER.error(
                "The config at %s could not be interpreted as a Python module.", configpath
            )
            sys.exit(1)
        else:
            cfg = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(cfg)  # type: ignore[union-attr]

        try:
            # validate config
            config.validate()
        except RuntimeError as exc:
            LOGGER.error(exc)
            sys.exit(1)


config = Config()
"""This is the runtime configuration object. It is exposed on the module level via:
```python
from cobib.config import config
```
"""
config.defaults()



"""coBib's Author class."""

from __future__ import annotations

from typing import NamedTuple

from ruamel import yaml


class Author(NamedTuple):
    """A named tuple storing detailed author information.

    Whenever this object gets converted to a string (for printing/dumping/etc.) the following four
    possible cases may be the outcome:

    1. particle Last, suffix, First
    2. Last, suffix, First
    3. particle Last, First
    4. Last, First
    """

    first: str
    """The author's first name."""
    last: str
    """The author's last name."""
    particle: str | None = None
    """The author's name article."""
    suffix: str | None = None
    """The author's name suffix."""

    def __str__(self) -> str:
        """Formats this Author as a string."""
        if self.suffix is not None:
            if self.particle is not None:
                return f"{self.particle} {self.last}, {self.suffix}, {self.first}"
            return f"{self.last}, {self.suffix}, {self.first}"
        if self.particle is not None:
            return f"{self.particle} {self.last}, {self.first}"
        return f"{self.last}, {self.first}"

    @classmethod
    def to_yaml(
        cls, representer: yaml.representer.Representer, node: Author
    ) -> yaml.nodes.MappingNode:
        """Dumps the object for storage in YAML format.

        Args:
            representer: the YAML formatter object.
            node: the current author object to be formatted.

        Returns:
            A YAML-interpretable encoding of this author object.
        """
        return representer.represent_dict(
            {k: v for k, v in node._asdict().items() if v is not None}
        )

    @classmethod
    def parse(cls, author: str) -> str | Author:
        """Parses a string and extracts the author information.

        If the string is fully enclosed by curly braces, it will not be processed further. This can
        be used to prevent the splitting of (for example) the names of non-individuals (such as
        companies).

        Otherwise, the following forms are understood:

        1. First von Last
        2. von Last, First
        3. von Last, Jr, First

        The cases above are distinguished based on the number of commas that are found in the
        string.

        Args:
            author: the string to be parsed.

        Returns:
            The parsed author information (or the verbatim author string).

        Raises:
            ValueError: when more then 2 commas are encountered in the input string.
        """
        if author[0] == "{" and author[-1] == "}":
            return author

        num_commas = author.count(",")

        if num_commas == 0:
            parts = author.split()
            if len(parts) == 1:
                return author

            last = parts[-1]
            first = parts[0]
            particle = None
            for part in parts[1:-1]:
                if part[0].islower():
                    if particle:
                        particle += " " + part  # type: ignore[unreachable]
                    else:
                        particle = part
                else:
                    first += " " + part

            return Author(first, last, particle=particle)

        if num_commas == 1:
            prefixed_last, first = author.split(",")
            suffix = None
        elif num_commas == 2:  # noqa: PLR2004
            prefixed_last, suffix, first = author.split(",")
            suffix = suffix.strip()
        else:
            raise ValueError(f"Expected less than 3 commas in the input string, not {num_commas}.")

        split_last = prefixed_last.split()
        if len(split_last) == 1:
            last = split_last[0]
            particle = None
        else:
            last_ = []
            particle_ = []
            for part in split_last:
                if part[0].islower():
                    particle_.append(part)
                else:
                    last_.append(part)
            last = " ".join(last_)
            particle = " ".join(particle_) or None

        return Author(first.strip(), last, particle=particle, suffix=suffix)



"""coBib's Git command.

This command is a simple pass through to the `git` executable, running inside the folder of coBib's
database (i.e. it is equivalent to running `git -C <path/to/cobib/database> <whatever arguments>`).
It exists solely for convenience since the above can be cumbersome to type out and specifying a
shell alias only works when coBib is used for a single database location (otherwise one needs one
alias for each database).

.. note::

   This command is not available from within the TUI because the generated output is not dealt with
   very nicely (yet? 🤔)

Below are a few example use cases.

### Checking the latest change
```
cobib git show HEAD
```

### Checking for uncommitted changes
```
cobib git status
```

### Browsing the entire history
```
cobib git log
```

### Pushing or pulling the recent changes
```
cobib git push origin master
cobib git pull origin master
```

### This also works with coBib-level keyword arguments:

```
cobib -c my_other_config.py git show HEAD
cobib -l my_log_file.txt git log
```

<hr/>

Of course, you have the full power of git at your fingertips, so there really are no limits to what
you can do.
"""

from __future__ import annotations

import argparse
import logging
import subprocess

from typing_extensions import override

from cobib.config import Event, config
from cobib.utils.git import is_inside_work_tree
from cobib.utils.rel_path import RelPath

from .base_command import Command

LOGGER = logging.getLogger(__name__)
"""@private module logger."""


class GitCommand(Command):
    """The Git Command.

    This command ignores all arguments passed to it and instead forwards them to the `git`
    executable for further processing.
    """

    name = "git"

    @override
    @classmethod
    def init_argparser(cls) -> None:
        parser = argparse.ArgumentParser(
            prog="git", description="Git subcommand parser.", exit_on_error=True
        )
        parser.add_argument("git_args", nargs=argparse.REMAINDER, help="the arguments to git")
        cls.argparser = parser

    @override
    @classmethod
    def _parse_args(cls, args: tuple[str, ...]) -> argparse.Namespace:
        largs = super()._parse_args(())
        largs.git_args = args
        return largs

    @override
    def execute(self) -> None:
        git_tracked = config.database.git
        if not git_tracked:
            msg = (
                "You must enable coBib's git-tracking in order to use the `Git` command."
                "\nPlease refer to the documentation for more information on how to do so."
            )
            LOGGER.error(msg)
            return

        file = RelPath(config.database.file).path
        root = file.parent
        if not is_inside_work_tree(root):
            msg = (
                "You have configured, but not initialized coBib's git-tracking."
                "\nPlease consult `cobib init --help` for more information on how to do so."
            )
            LOGGER.error(msg)
            return

        LOGGER.debug("Starting Git command.")

        Event.PreGitCommand.fire(self)

        subprocess.run(["git", "-C", root, *self.largs.git_args], check=False)

        Event.PostGitCommand.fire(self)



#!/usr/bin/env python3
"""coBib main body."""

import asyncio

from cobib.ui.cli import CLI


async def main() -> None:
    """Main async executable.

    coBib's main function used to parse optional keyword arguments and subcommands.
    """
    await CLI().run()


def _main() -> None:
    """The main method wrapping the async method with `asyncio.run`."""
    asyncio.run(main())


if __name__ == "__main__":
    _main()  # pragma: no cover
